from argparse import ArgumentParser, RawDescriptionHelpFormatter
import base64
import json
from pathlib import Path
import pickle
import os
from tqdm import tqdm
import pandas as pd
from typing import Mapping
import plotly.express as px

from as_analytics.lsreport import read_filetree_ls

df = None


def _get_parser():
    parser = ArgumentParser(
        description="Parser for output of `tools/lstraverse.sh`.",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "path",
        type=str,
        help="Path of filtree as output from `tools/lstraverse.sh`.",
    )
    parser.add_argument(
        "--expid",
        type=str,
        help="Expid. Only for saving purposes",
    )
    parser.add_argument(
        "--root",
        type=str,
        default="",
        help="Root directory to examine.",
    )
    parser.add_argument(
        "--figure",
        type=str,
        default=None,
        choices=["treemap", "sunburst", "icicle", "all", None],
        help="Root directory to examine.",
    )
    parser.add_argument(
        "--show",
        action="store_true",
        help="Whether to show the figure(s) after plotting.",
    )
    parser.add_argument(
        "--keep-dots",
        action="store_true",
        help="Whether to keep the entries for '.' and '..' in the original dataframe.",
    )
    parser.add_argument("--cache", action="store_true", help="Read / write to ~/.cache")
    return parser


def _json_serialize(obj):
    if isinstance(obj, Path):
        return str(obj)
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")


def _get_cache_path(name, *args, **kwargs) -> Path:
    key = json.dumps([name, *args, kwargs], default=_json_serialize)
    encoded = base64.b64encode(key.encode("utf-8"))
    file = encoded.decode("utf-8")
    directory = os.path.basename(__file__)
    return Path.home().resolve() / ".cache" / directory / file


def _cache_read(cache, name, /, *args, **kwargs):
    if cache is False:
        return
    cpath = _get_cache_path(name, *args, **kwargs)
    if not cpath.exists():
        return None
    print(f"Reading '{name}' from cache...")
    with open(cpath, "rb") as fb:
        return pickle.load(fb)


def _cache_write(cache, obj, name, /, *args, **kwargs):
    if cache is False:
        return
    print(f"Writing '{name}' to cache...")
    cpath = _get_cache_path(name, *args, **kwargs)
    if not cpath.parent.exists():
        cpath.parent.mkdir(parents=True)
    with open(cpath, "wb") as fb:
        pickle.dump(obj, fb)


def _add_path_column(df):
    # Normalize Parent paths
    common_root = os.path.commonpath(list(df["Parent"]))
    common_root_parent = os.path.dirname(common_root)
    df["Parent"] = df["Parent"].str.removeprefix(common_root_parent)

    # Create full paths and normalize them
    df["Path"] = (df["Parent"] + "/" + df["Filename"]).apply(
        lambda x: x.removeprefix("/")
    )
    return df


def _get_subtree_df(df, root):
    # Normalize Parent paths
    df = df.copy()

    # Compute common root
    common_root = os.path.commonpath(list(df["Parent"]))

    # Normalize requested root
    if root in (None, ""):
        abs_root = common_root
    else:
        # If root is absolute, use it directly
        if os.path.isabs(root):
            abs_root = os.path.normpath(root)
        else:
            # If root is relative, join it with common root
            abs_root = os.path.normpath(os.path.join(common_root, root))

    # Filter to subtree - only include files that are under the absolute root
    mask = df["Parent"].str.startswith(abs_root)
    subtree_df = df[mask].copy()

    if subtree_df.empty:
        raise ValueError(
            f"No entries found under root '{root}' (resolved to '{abs_root}')."
        )

    subtree_df["name"] = subtree_df["Filename"]

    return subtree_df


def _human_readable_size(size_bytes):
    """Convert size in bytes to human readable format."""
    for unit in ["B", "K", "M", "G", "T"]:
        if size_bytes < 1024:
            return f"{size_bytes:.0f}{unit}"
        size_bytes /= 1024
    return f"{size_bytes:.0f}P"


def draw(
    sizes: Mapping[str, int],
    parents: Mapping[str, str],
    root: str = None,
    figure: str = "treemap",
):
    """
    Visualize file structure under a given root directory using Plotly.

    Parameters:
        sizes (Mapping[str, int]): Dictionary mapping directory paths to their total sizes
        parents (Mapping[str, str]): Dictionary mapping directory paths to their parent directories
        root (str | None): Subdirectory to start visualization from. Can be absolute or relative to the common root.
        figure (str): One of 'treemap', 'sunburst', or 'icicle'.
    """
    if figure not in {"treemap", "sunburst", "icicle"}:
        raise ValueError(
            f"Unsupported figure '{figure}'. Use 'treemap', 'sunburst', or 'icicle'."
        )

    # Create visualization DataFrame
    viz_df = pd.DataFrame(
        {
            "id": list(sizes.keys()),  # Full path as ID
            "parent": list(parents.values()),  # Full parent path
            "value": list(sizes.values()),  # Directory size
        }
    )

    # Add display name (last part of path + size)
    viz_df["name"] = viz_df.apply(
        lambda row: f"{row['id'].split('/')[-1]} ({_human_readable_size(row['value'])})",
        axis=1,
    )

    # Add human readable size for hover
    viz_df["size_str"] = viz_df["value"].apply(_human_readable_size)

    # Filter by root if specified
    if root:
        mask = viz_df["id"].str.startswith(root)
        viz_df = viz_df[mask]
        if viz_df.empty:
            raise ValueError(f"No entries found under root '{root}'.")

    # Create the figure
    fig = getattr(px, figure)(
        viz_df,
        ids="id",  # Use full paths as IDs
        names="name",  # Show name with size
        parents="parent",  # Use full parent paths
        values="value",
        hover_data=[
            "id",
            "size_str",
        ],  # Show full path and human readable size in hover
    )

    # Get the root for the title
    abs_root = root if root else viz_df[viz_df["parent"] == ""]["id"].iloc[0]

    # Update traces for better visibility
    fig.update_traces(
        root_color="lightgrey",
        textposition="middle center",  # Center text in each box
        textfont=dict(size=14),  # Larger text
        textinfo="label",  # Show the label (name) in each box
    )

    # Update layout for better visibility and larger size
    fig.update_layout(
        title=dict(
            text=f"{figure.title()} of directory sizes under: {abs_root}",
            font=dict(size=24),  # Larger title
        ),
        margin=dict(t=60, l=0, r=0, b=0),  # More space for title
        width=2000,  # Wider figure
        height=1500,  # Taller figure
        font=dict(size=14),  # Larger font for all text
    )

    return fig


def makefigure(sizes, parents, root, figure, show, save, prefix, cache, **kwargs):

    if figure == "all":
        return [
            makefigure(
                sizes, parents, root, "treemap", show, save, prefix, cache, **kwargs
            ),
            makefigure(
                sizes, parents, root, "sunburst", show, save, prefix, cache, **kwargs
            ),
            makefigure(
                sizes, parents, root, "icicle", show, save, prefix, cache, **kwargs
            ),
        ]

    print(f"Plotting {figure}.")
    fig = _cache_read(cache, "figure", root=root, figure=figure, **kwargs)
    if fig is None:
        fig = draw(sizes, parents, root, figure)
        _cache_write(cache, fig, "figure", root=root, figure=figure, **kwargs)
    if save:
        fig.write_image(f"figures/{prefix}{figure}.png")
    if show:
        fig.show()
    return fig


def compute_directory_size(df):
    """
    Compute directory sizes and their hierarchical relationships.

    Parameters:
        df (pd.DataFrame): DataFrame with 'Parent' and 'Filename' columns

    Returns:
        tuple: (sizes, parents) where:
            - sizes: dict mapping directory paths to their total sizes
            - parents: dict mapping directory paths to their parent directories
    """
    # Initialize dictionaries
    direct_sizes = {}  # Size of files directly in each directory
    total_sizes = {}  # Total size including subdirectories
    parents = {}

    # First pass: compute direct sizes and parent relationships
    for directory in tqdm(sorted(df["Parent"].unique())):
        # Get files directly in this directory
        mask = df["Parent"] == directory
        direct_sizes[directory] = int(df.loc[mask, "Filesize"].sum())
        total_sizes[directory] = direct_sizes[directory]  # Initialize total size

        # Set parent relationship
        if "/" not in directory:
            parents[directory] = ""
        else:
            parent_dir, _ = directory.rsplit("/", 1)
            parents[directory] = parent_dir if parent_dir in direct_sizes else ""

    # Second pass: propagate sizes up the tree
    # Process directories from longest to shortest path
    for directory in sorted(direct_sizes.keys(), key=len, reverse=True):
        parent = parents[directory]
        if parent:  # If not root
            total_sizes[parent] += total_sizes[directory]

    return total_sizes, parents


def main():
    global df
    args = _get_parser().parse_args()
    path = Path(args.path).resolve()
    expid = args.expid
    root = args.root
    figure = args.figure
    show = args.show
    keep_dots = args.keep_dots
    cache = args.cache

    df = None

    df = _cache_read(cache, "df", path=path, expid=expid, keep_dots=keep_dots)
    if df is None:
        df = read_filetree_ls(path, join=True, common=False, keep_dots=keep_dots)
        _cache_write(cache, df, "df", path=path, expid=expid, keep_dots=keep_dots)

    df = _add_path_column(df)
    df = _get_subtree_df(df, root)

    sizes_parents = _cache_read(
        cache, "sizes+parents", path=path, expid=expid, keep_dots=keep_dots, root=root
    )
    if sizes_parents is None:
        sizes, parents = compute_directory_size(df)
        _cache_write(
            cache,
            (sizes, parents),
            "sizes+parents",
            path=path,
            expid=expid,
            keep_dots=keep_dots,
            root=root,
        )

    if figure is not None:
        makefigure(
            sizes=sizes_parents[0],
            parents=sizes_parents[1],
            root=root,
            figure=figure,
            show=show,
            save=True,
            prefix=(f"{expid}_" if expid else ""),
            cache=cache,
            path=path,
            expid=expid,
            keep_dots=keep_dots,
        )


if __name__ == "__main__":
    main()
